//oop_ex130.cpp
//class templates and non-type parameters

#include <iostream>
#include "myString.h"

using namespace std;

//---class template---
template<class T, int id>
class List
{
	struct Node
	{
		Node* prev;
		Node* suc;
		T val;

		Node(Node* p, Node* s, const T& v)
			: prev(p), suc(s), val(v) { }
	};

	Node* head;

public:
	List() :head(0) {}

	List(const T& t) : head(new Node(0, 0, t)) {}

	void print_all()
	{
		cout << "List " << id << " : ";
		for (Node* p = head; p; p = p->suc)
			cout << p->val << " ";
		cout << endl;
	}

	void add_to_list(List<T, id>&, T);
	void remove_from_list(List<T, id>&, T);
};


//---function templates---
template<class T, int id>
void List<T, id>::add_to_list(List<T, id>& li, T data)
{
	Node* new_Node = new Node(0, li.head, data);
	li.head = new_Node;
	new_Node->suc->prev = head;
}

template<class T, int id>
void List<T, id>::remove_from_list(List<T, id>& li, T data)
{
	Node* p;
	for (p = li.head; p->val != data; p = p->suc)
	{
		if (p->suc == 0)
			return;
	}

	if (p == li.head)	//if item to be deleted is at the head
	{
		li.head = li.head->suc;
		li.head->prev = 0;
		delete p;
		return;
	}

	p->prev->suc = p->suc;
	if (p->suc != 0)	//if item to be deleted is not at the end
		p->suc->prev = p->prev;
	delete p;
}

void main()
{
	// use the template class for a list of int

	int i = 100;
	List<int, 1>* numlist = new List<int, 1>(i);
	numlist->print_all();

	cout << "adding 500 to the list" << endl;
	numlist->add_to_list(*numlist, 500);
	numlist->print_all();

	cout << "adding 600 to the list" << endl;
	numlist->add_to_list(*numlist, 600);
	numlist->print_all();


	cout << "removing 100 from the list" << endl;
	numlist->remove_from_list(*numlist, 100);
	numlist->print_all();

	cout << "removing 600 from the list" << endl;
	numlist->remove_from_list(*numlist, 600);
	numlist->print_all();

	cout << "removing 700 from the list" << endl;
	numlist->remove_from_list(*numlist, 700);
	numlist->print_all();


	// use the template class for a list of String

	String str = "red";
	List<String, 2>* stringlist = new List<String, 2>(str);
	stringlist->print_all();

	cout << "adding green to the list" << endl;
	stringlist->add_to_list(*stringlist, "green");
	stringlist->print_all();

	cout << "adding blue to the list" << endl;
	stringlist->add_to_list(*stringlist, "blue");
	stringlist->print_all();


	cout << "removing red from the list" << endl;
	stringlist->remove_from_list(*stringlist, "red");
	stringlist->print_all();

	cout << "removing blue from the list" << endl;
	stringlist->remove_from_list(*stringlist, "blue");
	stringlist->print_all();

	cout << "removing yellow from the list" << endl;
	stringlist->remove_from_list(*stringlist, "yellow");
	stringlist->print_all();

	system("pause");
}